import csv
import logging
import math
from typing import TextIO, Dict, List


def write_freq_file(out_filename: str, freq_dict: Dict[str, int], max_records=-1):
    """Writes the given frequency table to out_filename, in order of descending value.

    Stops at max_records if this optional argument is positive"""
    freq_items = sorted(freq_dict.items(), key=lambda x: x[1], reverse=True)
    rank = 1
    with open(out_filename, 'w') as out_fh:
        logging.debug(f'Writing frequencies to {out_filename}')
        writer = csv.writer(out_fh)
        for freq_item in freq_items:
            writer.writerow(freq_item)
            rank += 1
            if rank > max_records > 0:
                break
        logging.debug(f'Wrote {len(freq_items)} entries.')


def write_rank_file(out_filename: str, ranked_items: List[str], max_records=-1):
    """Writes the elements of the given list to out_filename, in order.

    Stops at max_records if this optional argument is positive,"""
    if max_records > 0:
        ranked_items = ranked_items[:max_records]
    with open(out_filename, 'w') as out_fh:
        out_fh.write("\n".join(ranked_items))
        logging.debug(f'Wrote {len(ranked_items)} entries to {out_filename}.')


def read_freq_file(input_fh: TextIO, max_records=-1) -> Dict[str, int]:
    """Reads lines of the form "word, count" and returns a dictionary of this word -> count data

    If there are duplicate keys, the corresponding values are added.
    """
    output_dict = {}
    reader = csv.reader(input_fh)
    lines_read = 1
    for row in reader:
        output_dict[row[0]] = output_dict.get(row[0], 0) + int(row[1])
        lines_read += 1
        if lines_read > max_records > 0:
            break
    return output_dict


def read_rank_file(input_fh: TextIO, max_records=-1) -> Dict[str, int]:
    """Reads lines and returns a dictionary where each key is a line and each value is the rank / line number.

    Repeated lines are ignored.
    """
    output_ranks = {}
    line_number = 1
    for line in input_fh:
        text = line.strip()
        if text not in output_ranks:
            output_ranks[text] = line_number
        line_number += 1
        if line_number > max_records > 0:
            break
    return output_ranks


def freq_dict_to_lowercase(input_dict: Dict[str, int]) -> Dict[str, int]:
    """Takes a dictionary of string to value, and returns one where the keys are all mapped to
    lowercase and the values are the sums of all the values mapped in this way."""
    output_dict = {}
    for item in input_dict.items():
        lc = item[0].lower()
        output_dict[lc] = output_dict.get(lc, 0) + item[1]
    return output_dict


def freq_table_to_ranks(input_dict: Dict[str, int]):
    return {item[0]: rank + 1 for rank, item in enumerate(sorted(input_dict.items(), key=lambda x: x[1], reverse=True))}


def normalize_score_dict(input_dict: Dict[str, float]) -> Dict[str, float]:
    """Takes a dictionary of scores and applies L1-normalization (dividing each value by the sum).

    This is the simplest way of turning a collection of scores into a probability distribution.
    """
    total_weight = sum(input_dict.values())
    output_dict = {key: value / total_weight if total_weight > 0 else 1 / len(input_dict)
                   for key, value in input_dict.items()}
    return output_dict


def softmax_score_dict(input_dict: Dict[str, float]) -> Dict[str, float]:
    """Takes a dictionary of scores and applies the softmax function.

    This produces a probability distribution where differences in the inputs are made more pronounced.
    """
    total_weight = sum([math.exp(x) for x in input_dict.values()])
    output_dict = {key: math.exp(value) / total_weight for key, value in input_dict.items()}
    return output_dict


def softmax_to_l1(input_dict: Dict[str, float]) -> Dict[str, float]:
    """Takes a dictionary of scores and renormalizes to the probability distribution of which the input would
    be the softmax. (Doesn't check that the input is actually probability distribution, i.e., values add to one.

    If the softmax was created from a probability distribution, this will recover that distribution.
    """
    log_dict = {key: math.log(value) for key, value in input_dict.items()}
    sum_logs = sum(log_dict.values())
    remainder_shared = (1 - sum_logs) / len(input_dict)
    return {key: value + remainder_shared for key, value in log_dict.items()}


def precision_recall_f(total_positives, attempted, correct, beta=1):
    """Returns the precision, recall, and f1-score given with this number of
    total positives, attempted, and correct items."""
    precision = correct / attempted
    recall = correct / total_positives
    f_measure = (1 + beta * beta) * precision * recall / (beta * beta * precision + recall)
    return precision, recall, f_measure



from lplangid import tokenizer


def test_tokenize_fast():
    assert tokenizer.tokenize_fast("Hello World!") == ["Hello", "World"]
    assert tokenizer.tokenize_fast("U.S.A.") == ["U.S.A"]
    assert tokenizer.tokenize_fast("U.S.A. means United States") == ["U.S.A", "means", "United", "States"]
    assert tokenizer.tokenize_fast("John's 'socks'") == ["John's", "socks"]
    assert tokenizer.tokenize_fast("Text with <br> HTML linebreak") == ["Text", "with", "HTML", "linebreak"]


def test_tokenize_url():
    assert tokenizer.tokenize_fast("http://go.microsoft.com/fwlink/?LinkID=267510&clcid=0x409") \
           == ["http://go.microsoft.com/fwlink/?LinkID=267510&clcid=0x409"]


def test_remove_html():
    assert tokenizer.remove_html_tags("Hi <br> there") == "Hi  there"
    assert tokenizer.remove_html_tags("Oh dear < This will all be removed >") == "Oh dear "



"""
This file contains data preparation / utilities.
It is mainly used when adding new languages to the lplangid classifier.

See the README.md file in this directory for more detailed instructions.
"""

import argparse
import logging
import math
import os
import sys
from typing import Dict, List, Tuple

from training.data_overrides import TOP_DATA_OVERRIDES, RANKED_DATA_OVERRIDES
from lplangid import language_classifier as lc, count_utils as cu
from lplangid.tokenizer import tokenize_fast as tokenize

"""This should be set to wherever the Wikipedia files are on your system."""
WIKI_TEXT_ROOT = os.path.expanduser('~/Data/Wikipedia/')

SKIP_WORDS_WITH_DIGITS = True
MIN_WORD_LENGTH = 2
MAX_NUM_WORDS_PER_LANGUAGE = 5000

NEW_FREQ_DATA = os.path.join(os.path.dirname(os.path.abspath(__file__)), "new_freq_data")
ALL_WORDS = -1  # triggers use of all words
ALL_CHARS = -1  # triggers use of all chars
ALL_FILES = -1  # triggers use of all files


def train_test_split(requested_languages=()):
    """Processes the output of WikiExtractor.py (see https://github.com/attardi/wikiextractor) into train and test.

    This is potentially brittle: you should only do this once when adding a new language or updating everything.

    :param requested_languages: List of languages to process. If empty, all available on filesystem will be processed.
    """
    test_proportion = 0.2
    lang_dirs = [lang for lang in os.listdir(WIKI_TEXT_ROOT) if len(lang) == 2]
    if requested_languages:
        missing_languages = [ml for ml in requested_languages if ml not in lang_dirs]
        if missing_languages:
            raise ValueError(
                f"Language '{missing_languages}' was requested but there is no such directory in {WIKI_TEXT_ROOT}")
        lang_dirs = requested_languages
    logging.info(f"Making test train split for languages: {', '.join(lang_dirs)}")

    os.chdir(WIKI_TEXT_ROOT)
    for lang in lang_dirs:
        logging.info(f"\tMaking test train split for language: {lang}")
        os.chdir(lang)
        lang_dir = os.getcwd()
        text_dirs = [x for x in os.listdir(lang_dir) if os.path.isdir(x)]
        if text_dirs != ["text"]:
            logging.info(f"\tExpected just 'text' directory in {os.getcwd()} - instead got {text_dirs}. "
                         f"Skipping the rest of test / train split for language {lang}.")
            continue
        os.mkdir("train")
        os.mkdir("test")
        for subdir in os.listdir("text"):
            os.mkdir(os.path.join("train", subdir))
            os.mkdir(os.path.join("test", subdir))
            all_text_files = sorted(os.listdir(os.path.join("text", subdir)))
            if len(all_text_files) == 0:
                continue
            cutoff = math.floor(len(all_text_files) * (1 - test_proportion))
            train_files = all_text_files[:cutoff]
            for train_file in train_files:
                os.rename(os.path.join("text", subdir, train_file),
                          os.path.join("train", subdir, train_file))
            test_files = all_text_files[cutoff:]
            for testfile in test_files:
                os.rename(os.path.join("text", subdir, testfile),
                          os.path.join("test", subdir, testfile))
            os.rmdir(os.path.join("text", subdir))
        os.rmdir("text")
        os.chdir("../..")


def count_from_text_root(wiki_text_dir):
    """Counts the word and character frequencies in text files beneath the given wiki_text_dir.

    Words are normalized to lowercase, but upper and lowercase characters are counted separately.
    """
    term_freq_dict = {}
    char_freq_dict = {}
    for path, _, filenames in os.walk(wiki_text_dir):
        for filename in filenames:
            filepath = os.path.join(path, filename)
            with open(filepath) as filehandle:
                for line in filehandle:
                    if line.startswith('<'):
                        continue
                    for char in [char for char in line if char.isalpha()]:
                        if char.isalpha():
                            char_freq_dict[char] = char_freq_dict.get(char, 0) + 1
                    for word in tokenize(line):
                        if len(word) < MIN_WORD_LENGTH:
                            continue
                        if SKIP_WORDS_WITH_DIGITS and any([x.isdigit() for x in word]):
                            continue
                        term_freq_dict[word.lower()] = term_freq_dict.get(word.lower(), 0) + 1
    return term_freq_dict, char_freq_dict


def count_wiki_text_contents(requested_languages: Tuple[str] = ()):
    """
    For each wiki directory (identified by a 2-character language code), counts the words and characters
    and outputs to appropriate rank files (for words) and frequency files (for characters).

    Each wiki directory must have a "train" or a "text" directory with the text files to be counted.

    :param requested_languages: List of languages to process. If empty, all available on filesystem will be processed.
    """
    lang_dirs = [lang for lang in os.listdir(WIKI_TEXT_ROOT) if len(lang) == 2]
    if requested_languages:
        missing_languages = [ml for ml in requested_languages if ml not in lang_dirs]
        if missing_languages:
            raise ValueError(
                f"Language '{missing_languages}' was requested but there is no such directory in {WIKI_TEXT_ROOT}")
        lang_dirs = requested_languages
    logging.info(f"Making count resources for languages: {', '.join(lang_dirs)}")

    for lang in lang_dirs:
        logging.info(f"\tStarting term and character counting for language '{lang}' ...")
        lang_dir_full = os.path.join(WIKI_TEXT_ROOT, lang)
        lang_dir_contents = os.listdir(lang_dir_full)
        text_dir = "train" if "train" in lang_dir_contents else "text" if "text" in lang_dir_contents else None
        if not text_dir:
            raise ValueError(f"No 'train' or 'text' directory in {lang_dir_full}. Please investigate. "
                             f"Check that Wiki archive was uncompressed using bunzip2 "
                             f"and text extracted using WikiExtractor.")

        term_freq_dict, char_freq_dict = count_from_text_root(os.path.join(lang_dir_full, text_dir))
        term_rank_list = [kv[0] for kv in sorted(term_freq_dict.items(), key=lambda x: x[1], reverse=True)]

        # Write out word in ranked order. The "+ 1000" is in case some are filtered out later.
        cu.write_rank_file(os.path.join(
            lc.FREQ_DATA_DIR, f"{lang}_term_rank.csv"), term_rank_list, max_records=MAX_NUM_WORDS_PER_LANGUAGE+1000)
        cu.write_freq_file(os.path.join(
            lc.FREQ_DATA_DIR, f"{lang}_char_freq.csv"), char_freq_dict)
        logging.info(f"\tFinished counting and writing wiki word ranks and character frequencies for language '{lang}'"
                     f"into directory {lc.FREQ_DATA_DIR}.")


def get_char_lag(all_char_ranks: Dict[str, List[Tuple[str, float]]], lang: str, term: str):
    """Returns a score comparing the given term's char score for the given language
    to the best char score for that term in any language."""
    term_scores = lc.score_chars(all_char_ranks, term)
    if lang not in term_scores or len(term_scores) == 0 or max(term_scores.values()) == 0:
        return 0
    return term_scores[lang] / max(term_scores.values())


def filter_terms_by_chars(requested_languages=()):
    """Rewrites each language's term rank file with only terms whose relative char score is above a threshold.

    :param requested_languages: List of languages to process. If empty, all available in freq_data will be processed.
    """
    char_lag_threshold = 0.2  # Determined empirically by staring at results.

    all_term_ranks, all_char_freqs = lc.prepare_scoring_tables()
    if not requested_languages:
        requested_languages = all_term_ranks.keys()
    for lang in requested_languages:
        term_ranks = all_term_ranks[lang]
        filtered_term_ranks = {}
        for term in term_ranks:
            if get_char_lag(all_char_freqs, lang, term) > char_lag_threshold:
                filtered_term_ranks[term] = term_ranks[term]
        ranked_terms = [kv[0] for kv in sorted(filtered_term_ranks.items(), key=lambda x: x[1])]
        cu.write_rank_file(os.path.join(
            lc.FREQ_DATA_DIR, f"{lang}_term_rank.csv"), ranked_terms, max_records=MAX_NUM_WORDS_PER_LANGUAGE)
        logging.info(f"Rewrote {lang}_term_rank.csv after filtering out {len(term_ranks) - len(ranked_terms)} "
                     f"out of {len(term_ranks)} terms for having unusual characters for language {lang}.")


def add_data_overrides_to_term_ranks(requested_languages=()):
    """Adds data from the data_overrides files to the term rank files in-place.

    Tries to avoid doing this twice by accident, though this is clumsy and should be checked by developers.

    :param requested_languages: List of languages to process. If empty, all available in freq_data will be processed.
    """
    all_term_ranks, _ = lc.prepare_scoring_tables()
    if not requested_languages:
        requested_languages = all_term_ranks.keys()
    for lang in requested_languages:
        if lang not in RANKED_DATA_OVERRIDES and lang not in TOP_DATA_OVERRIDES:
            logging.info(f"There are no listed data_overrides for language {lang}. Skipping this step.")
            continue

        old_ranks = all_term_ranks[lang]
        # Turn the rank dictionary into a list for easier manipulation rather than fast lookup.
        term_list = [x[0] for x in sorted(old_ranks.items(), key=lambda kv: kv[1])]
        if lang in RANKED_DATA_OVERRIDES:
            if not all([term in old_ranks for term in RANKED_DATA_OVERRIDES[lang]]):
                for new_term, rank in sorted(RANKED_DATA_OVERRIDES[lang].items(), key=lambda kv: kv[1], reverse=True):
                    term_list.insert(rank, new_term)
            else:
                logging.info(f"It is likely that ranked overrides were already added for language {lang}. Skipping.")

        if lang in TOP_DATA_OVERRIDES:
            if term_list[0] != TOP_DATA_OVERRIDES[lang][0]:
                term_list = TOP_DATA_OVERRIDES[lang] + term_list
            else:
                logging.info(f"It is likely that ranked overrides were already added for language {lang}. Skipping.")

        cu.write_rank_file(os.path.join(lc.FREQ_DATA_DIR, f"{lang}_term_rank.csv"), term_list)
        logging.info(f"Rewrote {lang}_term_rank.csv after adding data overrides.")


def main(argv):
    """Main function that takes a directory of text contents (e.g., from an uncompressed wiki archive), and
    - Splits the text files into train and test directories.
    - Counts the terms and characters in each language directory.
    - Filters out terms that don't fit the character profile of a language.
    - Adds known extra terms (data_overrides) to the term rank tables.
    """
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("--languages", help="Comma separated list of language codes, e.g., --languages=en,es,pt",
                        type=str)

    args = parser.parse_args(argv)
    requested_languages = args.languages.split(',') if args.languages else None

    logging.info("Starting processing wiki files.")
    if requested_languages:
        logging.info(f"Explicitly requested these languages: {', '.join(requested_languages)}")

    train_test_split(requested_languages=requested_languages)
    count_wiki_text_contents(requested_languages=requested_languages)
    filter_terms_by_chars(requested_languages=requested_languages)
    add_data_overrides_to_term_ranks(requested_languages=requested_languages)


if __name__ == '__main__':
    main(sys.argv[1:])



import logging
import math
import os
import string
from collections import Counter
from typing import Dict, List, Optional, Tuple

from lplangid import count_utils as cu
from lplangid.const_data import COMPUTERESE_STARTS
from lplangid.tokenizer import tokenize_fast

# The default Wikipedia + overrides datafiles are shipped in this base directory.
FREQ_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "freq_data")

# Weights and biases that can easily be tweaked by hand, or could be learned / optimized.
TERM_PRESENCE_WEIGHT = 0.05
BASELINE_TERM_SCORE = TERM_PRESENCE_WEIGHT / 2
TOP_RANK_DAMPING = 10
LETTERS = set(string.ascii_lowercase)

# If the character score for a language is below this proportion of the winning char score, it's rejected.
# This is tuned towards rejecting Chinese as an option if there are too many Japanese-only characters.
# (In other words, beware, this *is* a dangerously over-tuned hand-written rule that might not generalize.)
CHAR_MIN_TO_PLAY = 0.6

# Maximum number of words read in for each language.
MAX_WORDS_PER_LANG = 10000

CLASSIFY_WORDS_LOWER_CASE = True
CLASSIFY_CHARS_LOWER_CASE = True


class RRCLanguageClassifier:
    """RRCLanguageClassifier is a class that provides language detection scores and predictions.

    It holds the term rank and char rank tables, and runs the pure functions (below) using this state.
    """
    def __init__(self, term_ranks: Dict[str, Dict[str, int]], char_weights: Dict[str, List[Tuple[str, float]]]):
        """
         Construct a new 'RRCLanguageClassifier' object.

         :param term_ranks: dictionary mapping language code -> word/term -> rank.
         :param char_weights: dictionary mapping character -> (language, relative frequency).

         The char_weights table is optimized to score every (character, language) score, whereas the term_ranks
         table is optimized to compute (language, term) scores for languages that pass the character cutoff.
         """
        self.term_ranks: Dict[str, Dict[str, int]] = term_ranks
        self.char_weights: Dict[str, List[Tuple[str, float]]] = char_weights

    @staticmethod
    def default_instance():
        """Gets a default instance populated using the prepare_scoring_tables function."""
        all_term_ranks, all_char_weights = prepare_scoring_tables()
        logging.info(f"Loaded classifier with term ranks and character frequencies for these languages: "
                     f"{', '.join(sorted(all_term_ranks.keys()))}")
        return RRCLanguageClassifier(all_term_ranks, all_char_weights)

    def get_winner(self, text: str) -> str:
        """Returns the language with the single best score. (Ties are very rare.)"""
        return get_winner(self.term_ranks, self.char_weights, text)

    def get_winner_score(self, text: str) -> Tuple[str, float]:
        """Returns the language with the single best score, and its score. (Ties are very rare.)"""
        return get_winner_score(self.term_ranks, self.char_weights, text)

    def get_language_scores(self, text: str) -> List[Tuple[str, float]]:
        """Returns a list of (language code, score) pairs, sorted from highest to lowest score."""
        return score_text(self.term_ranks, self.char_weights, text)


def prepare_scoring_tables() -> Tuple[Dict[str, Dict[str, int]], Dict[str, List[Tuple[str, float]]]]:
    """Reads in term and character ranking data from the files in FREQ_DATA_DIR"""
    all_term_ranks = {}
    all_char_freqs = {}
    for lang_code in set([x[:2] for x in os.listdir(FREQ_DATA_DIR) if x.endswith('.csv') and not x.startswith('.')]):
        tf_file = os.path.join(FREQ_DATA_DIR, f"{lang_code}_term_rank.csv")
        with open(tf_file) as term_freq_file:
            term_freqs = cu.read_rank_file(term_freq_file, MAX_WORDS_PER_LANG)
            all_term_ranks[lang_code] = term_freqs
        with open(os.path.join(FREQ_DATA_DIR, f'{lang_code}_char_freq.csv')) as char_freq_file:
            all_char_freqs[lang_code] = cu.normalize_score_dict(cu.read_freq_file(char_freq_file))
    all_char_weights = invert_char_tables(all_char_freqs)

    logging.debug(f"Prepared term and character ranking tables for languages: {sorted(all_term_ranks.keys())}")
    return all_term_ranks, all_char_weights


def invert_char_tables(lang_to_char_weight: Dict[str, Dict[str, float]]) -> Dict[str, List[Tuple[str, float]]]:
    """Inverts a table of lang -> char -> weight to a table of char -> list of (lang, weight) pairs."""
    all_char_weights_dict = {}
    for lang, char_weights in lang_to_char_weight.items():
        for char, weight in char_weights.items():
            if char not in all_char_weights_dict:
                all_char_weights_dict[char] = {}
            all_char_weights_dict[char][lang] = weight
    all_char_weights = {char: sorted(cu.normalize_score_dict(weights).items(), key=lambda x: x[1], reverse=True)
                        for char, weights in all_char_weights_dict.items()}
    return all_char_weights


def score_terms(all_term_ranks: Dict[str, Dict[str, int]], text: str, languages: Tuple[str] = ()) -> Dict[str, float]:
    """Gets a score for each language for the given text based on how common the terms are."""
    tokens = Counter(tokenize_fast(text))
    tokens = {token: count for token, count in tokens.items() if len(token) > 1 or token not in LETTERS}
    scores: Dict[str, float] = {}
    if not languages:
        languages = all_term_ranks.items()

    for lang in languages:
        lang_score = BASELINE_TERM_SCORE
        if lang in all_term_ranks:
            ranks = all_term_ranks[lang]
            for token, count in tokens.items():
                if token in ranks:
                    lang_score += (TERM_PRESENCE_WEIGHT + 1 / math.sqrt(TOP_RANK_DAMPING + ranks[token])) * count
        scores[lang] = lang_score
    return scores


def score_chars(all_char_weights: Dict[str, List[Tuple[str, float]]], text: str) -> Dict[str, float]:
    """Gets a score for each language for the given text based on how common the characters are."""
    chars = Counter(text)
    chars = {char: count for char, count in chars.items() if char.isalpha() and char in all_char_weights}
    scores = {}
    for char, count in chars.items():
        for lang, weight in all_char_weights[char]:
            if lang not in scores:
                scores[lang] = 0
            scores[lang] += weight * count
    return scores


def score_text(all_term_ranks: Dict[str, Dict[str, int]],
               all_char_weights: Dict[str, List[Tuple[str, float]]],
               text: str) -> List[Tuple[str, float]]:
    """Gets the term and character scores and combines them into a single score for each language."""
    if any([text.startswith(x) for x in COMPUTERESE_STARTS]):
        return []

    char_scores = score_chars(all_char_weights, text.lower() if CLASSIFY_CHARS_LOWER_CASE else text)
    if not any(char_scores):
        return []
    char_max = max(char_scores.values())
    char_scores = {k: v for k, v in char_scores.items() if v > char_max * CHAR_MIN_TO_PLAY}
    char_scores = cu.normalize_score_dict(char_scores) if sum(char_scores.values()) > 0 else char_scores
    # Early-out if there is only one contender left (partly to avoid penalizing no term matches without tokenization).
    if len(char_scores) == 1:
        return list(char_scores.items())

    term_scores = score_terms(
        all_term_ranks, text.lower() if CLASSIFY_WORDS_LOWER_CASE else text, languages=tuple(char_scores))
    # If we got this far but have no explicit term matches, then it's usually a spurious classification.
    if not max(term_scores.values()) >= BASELINE_TERM_SCORE + TERM_PRESENCE_WEIGHT:
        return []
    term_scores = cu.normalize_score_dict(term_scores)

    combined_scores = {lang: term_scores.get(lang, BASELINE_TERM_SCORE) * char_scores[lang] for lang in char_scores}
    return sorted(combined_scores.items(), key=lambda x: x[1], reverse=True)


def get_char_winner(all_char_weights: Dict[str, List[Tuple[str, float]]], text: str) -> str:
    """Calls score_chars and returns the language with the highest char score.
    If no scores are greater than zero, returns None."""
    scores = score_chars(all_char_weights, text)
    winner, score = max(scores.items(), key=lambda x: x[1])
    return winner if score > 0 else None


def get_winner_score(term_dict: Dict[str, Dict[str, int]], char_dict: Dict[str, List[Tuple[str, float]]], text: str
                     ) -> Tuple[Optional[str], float]:
    """Calls score_text and returns the winning language and its score.

    No thresholds or tie-breaking is used. If there is a tie, the winner is unpredictable.

    If all scores are zero, the winner is None."""
    combined_scores = score_text(term_dict, char_dict, text)
    if len(combined_scores) == 0:
        return None, 0
    winner, score = max(combined_scores, key=lambda x: x[1])
    return winner if score > 0 else None, score


def get_winner_margin(term_dict, char_dict, text) -> Tuple[Optional[str], float]:
    """Calls score_text and returns the winning language and how much it won by (compared with second highest score).

    If all scores are zero, the winner is None."""
    scores = score_text(term_dict, char_dict, text)
    sorted_scores: List[Tuple[str, float]] = sorted(scores, key=lambda x: x[1], reverse=True)
    if len(sorted_scores) == 0:
        return None, 0
    if len(sorted_scores) == 1:
        return sorted_scores[0]
    winner = sorted_scores[0][0] if sorted_scores[0][1] > 0 else None
    score = sorted_scores[0][1] - sorted_scores[1][1]
    return winner, score


def get_winner(all_term_ranks: Dict[str, Dict[str, int]],
               all_char_weights: Dict[str, List[Tuple[str, float]]],
               text: str) -> Optional[str]:
    """Calls score_text and returns the language with the highest score.
    If no scores are greater than zero, returns None."""
    combined_scores = score_text(all_term_ranks, all_char_weights, text)
    if len(combined_scores) == 0:
        return None
    winner, score = max(combined_scores, key=lambda x: x[1])
    return winner if score > 0 else None


def main():
    """Demonstration main function that you can run just to see what it does with a few phrases."""
    logging.basicConfig(level=logging.INFO)
    rrc_classifier = RRCLanguageClassifier.default_instance()
    texts = ["This is English", "Esto es español"]

    for text in texts:
        winner = rrc_classifier.get_winner(text)
        print(f'{winner}: {text}')


if __name__ == '__main__':
    main()



"""This file contains data used in classifier rules, e.g., do-not-classify system rules."""

"""Messages with these starting strings are never classified."""
COMPUTERESE_STARTS = {
    "<!", "EndToEndSystemTest", "Metadata", "[public/comment]", "displayText", "http"
}
